/*  eaton-pdu-marlin-mib.c - data to monitor Eaton ePDUs branded as:
 *                G2 Marlin SW / MI / MO / MA
 *                G3 Shark SW / MI / MO / MA
 *
 *  Copyright (C) 2008 - 2017
 * 		Arnaud Quette <arnaud.quette@gmail.com>
 * 		Arnaud Quette <ArnaudQuette@Eaton.com>
 *  Copyright (C) 2015 - 2017
 * 		Jim Klimov <EvgenyKlimov@Eaton.com>
 *
 *  Supported by Eaton <http://www.eaton.com>
 *   and previously MGE Office Protection Systems <http://www.mgeops.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "eaton-pdu-marlin-mib.h"
#include "dstate.h"

/* Eaton PDU-MIB - Marlin MIB
 * ************************** */

#define EATON_MARLIN_MIB_VERSION	"0.44"
#define EATON_MARLIN_SYSOID			".1.3.6.1.4.1.534.6.6.7"
#define EATON_MARLIN_OID_MODEL_NAME	".1.3.6.1.4.1.534.6.6.7.1.2.1.2.0"

static info_lkp_t marlin_outlet_status_info[] = {
	{ 0, "off" },
	{ 1, "on" },
	{ 2, "pendingOff" }, /* transitional status */
	{ 3, "pendingOn" },  /* transitional status */
	{ 0, NULL }
};

static info_lkp_t marlin_outletgroups_status_info[] = {
	{ 0, "off" },
	{ 1, "on" },
	{ 2, "rebooting" }, /* transitional status */
	{ 3, "mixed" },     /* transitional status, not sure what it means! */
	{ 0, NULL }
};

/* Ugly hack: having the matching OID present means that the outlet is
 * switchable. So, it should not require this value lookup */
static info_lkp_t marlin_outlet_switchability_info[] = {
	{ -1, "yes" },
	{ 0, "yes" },
	{ 0, NULL }
};

/* The physical type of outlet */
static info_lkp_t marlin_outlet_type_info[] = {
	{ 0, "unknown" },
	{ 1, "iecC13" },
	{ 2, "iecC19" },
	{ 10, "uk" },
	{ 11, "french" },
	{ 12, "schuko" },
	{ 20, "nema515" },
	{ 21, "nema51520" },
	{ 22, "nema520" },
	{ 23, "nemaL520" },
	{ 24, "nemaL530" },
	{ 25, "nema615" },
	{ 26, "nema620" },
	{ 27, "nemaL620" },
	{ 28, "nemaL630" },
	{ 29, "nemaL715" },
	{ 30, "rf203p277" },
	{ 0, NULL }
};

static info_lkp_t marlin_ambient_presence_info[] = {
	{ -1, "unknown" },
	{ 0, "no" },  /* disconnected */
	{ 1, "yes" }, /* connected */
	{ 0, NULL }
};

static info_lkp_t marlin_threshold_status_info[] = {
	{ 0, "good" },          /* No threshold triggered */
	{ 1, "warning-low" },   /* Warning low threshold triggered */
	{ 2, "critical-low" },  /* Critical low threshold triggered */
	{ 3, "warning-high" },  /* Warning high threshold triggered */
	{ 4, "critical-high" }, /* Critical high threshold triggered */
	{ 0, NULL }
};

static info_lkp_t marlin_threshold_frequency_status_info[] = {
	{ 0, "good" },          /* No threshold triggered */
	{ 1, "out-of-range" },  /* Frequency out of range triggered */
	{ 0, NULL }
};

static info_lkp_t marlin_ambient_drycontacts_info[] = {
	{ -1, "unknown" },
	{ 0, "open" },
	{ 1, "closed" },
	{ 0, NULL }
};

static info_lkp_t marlin_threshold_voltage_alarms_info[] = {
	{ 0, "" },                       /* No threshold triggered */
	{ 1, "low voltage warning!" },   /* Warning low threshold triggered */
	{ 2, "low voltage critical!" },  /* Critical low threshold triggered */
	{ 3, "high voltage warning!" },  /* Warning high threshold triggered */
	{ 4, "high voltage critical!" }, /* Critical high threshold triggered */
	{ 0, NULL }
};

static info_lkp_t marlin_threshold_current_alarms_info[] = {
	{ 0, "" },                       /* No threshold triggered */
	{ 1, "low current warning!" },   /* Warning low threshold triggered */
	{ 2, "low current critical!" },  /* Critical low threshold triggered */
	{ 3, "high current warning!" },  /* Warning high threshold triggered */
	{ 4, "high current critical!" }, /* Critical high threshold triggered */
	{ 0, NULL }
};

static info_lkp_t marlin_threshold_frequency_alarm_info[] = {
	{ 0, "" },                         /* No threshold triggered */
	{ 1, "frequency out of range!" },  /* Frequency out of range triggered */
	{ 0, NULL }
};

static info_lkp_t marlin_threshold_temperature_alarms_info[] = {
	{ 0, "" },                           /* No threshold triggered */
	{ 1, "low temperature warning!" },   /* Warning low threshold triggered */
	{ 2, "low temperature critical!" },  /* Critical low threshold triggered */
	{ 3, "high temperature warning!" },  /* Warning high threshold triggered */
	{ 4, "high temperature critical!" }, /* Critical high threshold triggered */
	{ 0, NULL }
};

static info_lkp_t marlin_threshold_humidity_alarms_info[] = {
	{ 0, "" },                        /* No threshold triggered */
	{ 1, "low humidity warning!" },   /* Warning low threshold triggered */
	{ 2, "low humidity critical!" },  /* Critical low threshold triggered */
	{ 3, "high humidity warning!" },  /* Warning high threshold triggered */
	{ 4, "high humidity critical!" }, /* Critical high threshold triggered */
	{ 0, NULL }
};

static info_lkp_t marlin_outlet_group_type_info[] = {
	{ 0, "unknown" },
	{ 1, "breaker1pole" },
	{ 2, "breaker2pole" },
	{ 3, "breaker3pole" },
	{ 4, "outlet-section" },
	{ 5, "user-defined" },
	{ 0, NULL }
};

static info_lkp_t marlin_input_type_info[] = {
	{ 1, "1" }, /* singlePhase     */
	{ 2, "2" }, /* splitPhase      */
	{ 3, "3" }, /* threePhaseDelta */
	{ 4, "3" }, /* threePhaseWye   */
	{ 0, NULL }
};

static char marlin_scratch_buf[20];

/* Compute the phase to which an outlet group is connected
 * WRT the number of phase(s) and the outlet group number.
 * Note that the group type (marlin_outlet_group_type_info) is
 *  not considered since this applies to any kind of group */
static const char *marlin_outlet_group_phase_fun(int outlet_group_nb)
{
	const char* str_phases_nb = dstate_getinfo("input.phases");
	int phases_nb = 1;
	if (str_phases_nb) {
		phases_nb = atoi(str_phases_nb);
		if (phases_nb == 1) {
			return "L1";
		}
		else { /* 3ph assumed, 2ph PDU don't exist! */
			if (outlet_group_nb > 3)
				snprintf(marlin_scratch_buf, 3, "L%i", (outlet_group_nb -3));
			else
				snprintf(marlin_scratch_buf, 3, "L%i", outlet_group_nb);

			return marlin_scratch_buf;
		}
	}
	return NULL;
}
static info_lkp_t marlin_outlet_group_phase_info[] = {
	{ 1, "dummy", marlin_outlet_group_phase_fun },
	{ 0, NULL }
};

/* Snmp2NUT lookup table for Eaton Marlin MIB */
static snmp_info_t eaton_marlin_mib[] = {

	/* Device collection */
	{ "device.mfr", ST_FLAG_STRING, SU_INFOSIZE, NULL, "EATON",
		SU_FLAG_STATIC | SU_FLAG_ABSENT | SU_FLAG_OK, NULL, NULL },
	{ "device.model", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.1.2.1.2.%i",
		"Eaton Powerware ePDU", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL },
	{ "device.serial", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.1.2.1.4.%i",
		"", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL },
	{ "device.type", ST_FLAG_STRING, SU_INFOSIZE, NULL, "pdu",
		SU_FLAG_STATIC | SU_FLAG_ABSENT | SU_FLAG_OK, NULL, NULL },
	{ "device.part", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.1.2.1.3.%i",
		"", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL },
	/* For daisychain, there is only 1 physical interface! */
	{ "device.macaddr", ST_FLAG_STRING, SU_INFOSIZE, ".1.3.6.1.2.1.2.2.1.6.2",
		"", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL },
	/* Daisychained devices support
	 * Notes: this definition is used to:
	 * - estimate the number of devices, based on the below OID iteration capabilities
	 * - determine the base index of the SNMP OID (ie 0 or 1) */
	{ "device.count", 0, 1, ".1.3.6.1.4.1.534.6.6.7.1.2.1.2.%i",
		"1", SU_FLAG_STATIC, NULL, NULL /* devices_count */ },

	/* UPS collection */
	{ "ups.mfr", ST_FLAG_STRING, SU_INFOSIZE, NULL, "EATON",
		SU_FLAG_STATIC | SU_FLAG_ABSENT | SU_FLAG_OK, NULL, NULL },
	{ "ups.model", ST_FLAG_STRING, SU_INFOSIZE,
		"1.3.6.1.4.1.534.6.6.7.1.2.1.2.%i",
		"Eaton Powerware ePDU", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL },

	/*	FIXME: use unitName.0	(ePDU)?
	 * { "ups.id", ST_FLAG_STRING, SU_INFOSIZE, AR_OID_DEVICE_NAME,
		"unknown", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL }, */
	{ "ups.serial", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.1.2.1.4.%i",
		"", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL },
	{ "ups.firmware", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.1.2.1.5.%i",
		"", SU_FLAG_STATIC | SU_FLAG_OK, NULL },
	{ "ups.type", ST_FLAG_STRING, SU_INFOSIZE, NULL, "pdu",
		SU_FLAG_STATIC | SU_FLAG_ABSENT | SU_FLAG_OK, NULL, NULL },
	 /* FIXME: needs a date reformating callback
	 *   2011-8-29,16:27:25.0,+1:0
	 *   Hex-STRING: 07 DB 08 1D 10 0C 36 00 2B 01 00 00
	 * { "ups.date", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.1.2.1.8.0",
		"", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL },
	 * { "ups.time", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.1.2.1.8.0",
		"", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL },
	 */

	/* Input collection */
	/* Historically, some of these data were previously published as
	 * outlet.{realpower,...}
	 * However, it's more suitable and logic to have these on input.{...}
	 */
	/* Note: the below gives the number of input, not the number of phase(s)! */
	/* inputCount.0; Value (Integer): 1
	{ "input.phases", 0, 1, ".1.3.6.1.4.1.534.6.6.7.1.2.1.20.0",
		NULL, SU_FLAG_STATIC | SU_FLAG_SETINT, NULL, &input_phases }, */
	/* Note: for daisychain mode, we must handle phase(s) per device, not as a whole */
	/* inputType.%i.1 = INTEGER: singlePhase (1) */
	{ "input.phases", 0, 1, ".1.3.6.1.4.1.534.6.6.7.3.1.1.2.%i.1",
		NULL, SU_FLAG_STATIC, &marlin_input_type_info[0], NULL },

	/* Frequency is measured globally */
	{ "input.frequency", 0, 0.1, ".1.3.6.1.4.1.534.6.6.7.3.1.1.3.%i.1",
		NULL, 0, NULL, NULL },
	{ "input.frequency.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.1.1.4.%i.1",
		NULL, SU_FLAG_OK, &marlin_threshold_frequency_status_info[0], NULL },
	{ "ups.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.1.1.4.%i.1",
		NULL, SU_FLAG_OK, &marlin_threshold_frequency_alarm_info[0], NULL },

	/* inputCurrentPercentLoad (measured globally)
	 * Current percent load, based on the rated current capacity */
	/* FIXME: input.load is mapped on input.L1.load for both single and 3phase !!! */
	{ "input.load", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.3.1.11.%i.1.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "input.L1.load", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.3.1.11.%i.1.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "input.L2.load", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.3.1.11.%i.1.2",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "input.L3.load", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.3.1.11.%i.1.3",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },

	/* FIXME:
	 * - Voltage is only measured per phase, as mV!
	 *   so input.voltage == input.L1.voltage for both single and 3phase
	 * - As per NUT namespace (http://www.networkupstools.org/docs/developer-guide.chunked/apas01.html#_valid_contexts)
	 *   Voltage has to be expressed either phase-phase or phase-neutral
	 *   This is depending on OID inputVoltageMeasType
	 *   INTEGER {singlePhase (1),phase1toN (2),phase2toN (3),phase3toN (4),phase1to2 (5),phase2to3 (6),phase3to1 (7)
	 * 		=> RFC input.Lx.voltage.context */
	{ "input.voltage", 0, 0.001, ".1.3.6.1.4.1.534.6.6.7.3.2.1.3.%i.1.1",
		NULL, 0, NULL, NULL },
	{ "input.voltage.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.4.%i.1.1",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "ups.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.4.%i.1.1",
		NULL, SU_FLAG_OK, &marlin_threshold_voltage_alarms_info[0], NULL },
	{ "input.voltage.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.5.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.voltage.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.6.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.voltage.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.7.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.voltage.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.8.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L1.voltage", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.3.%i.1.1",
		NULL, 0, NULL, NULL },
	{ "input.L1.voltage.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.4.%i.1.1",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "L1.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.4.%i.1.1",
		NULL, SU_FLAG_OK, &marlin_threshold_voltage_alarms_info[0], NULL },
	{ "input.L1.voltage.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.5.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L1.voltage.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.6.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L1.voltage.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.7.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L1.voltage.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.8.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L2.voltage", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.3.%i.1.2",
		NULL, 0, NULL, NULL },
	{ "input.L2.voltage.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.4.%i.1.2",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "L2.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.4.%i.1.2",
		NULL, SU_FLAG_OK, &marlin_threshold_voltage_alarms_info[0], NULL },
	{ "input.L2.voltage.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.5.%i.1.2",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L2.voltage.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.6.%i.1.2",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L2.voltage.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.7.%i.1.2",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L2.voltage.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.8.%i.1.2",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L3.voltage", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.3.%i.1.3",
		NULL, 0, NULL, NULL },
	{ "input.L3.voltage.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.4.%i.1.3",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "L3.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.4.%i.1.3",
		NULL, SU_FLAG_OK, &marlin_threshold_voltage_alarms_info[0], NULL },
	{ "input.L3.voltage.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.5.%i.1.3",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L3.voltage.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.6.%i.1.3",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L3.voltage.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.7.%i.1.3",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L3.voltage.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.2.1.8.%i.1.3",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	/* FIXME:
	 * - input.current is mapped on input.L1.current for both single and 3phase !!! */
	{ "input.current", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.4.%i.1.1",
		NULL, 0, NULL, NULL },
	{ "input.current.nominal", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.3.%i.1.1",
		NULL, 0, NULL, NULL },
	{ "input.current.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.5.%i.1.1",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "ups.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.5.%i.1.1",
		NULL, SU_FLAG_OK, &marlin_threshold_current_alarms_info[0], NULL },
	{ "input.current.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.6.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.current.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.7.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.current.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.8.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.current.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.9.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L1.current", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.4.0.1.1",
		NULL, 0, NULL, NULL },
	{ "input.L1.current.nominal", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.3.%i.1.1",
		NULL, 0, NULL, NULL },
	{ "input.L1.current.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.5.%i.1.1",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "L1.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.5.%i.1.1",
		NULL, SU_FLAG_OK, &marlin_threshold_current_alarms_info[0], NULL },
	{ "input.L1.current.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.6.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L1.current.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.7.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L1.current.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.8.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L1.current.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.9.%i.1.1",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L2.current", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.4.%i.1.2",
		NULL, 0, NULL, NULL },
	{ "input.L2.current.nominal", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.3.%i.1.2",
		NULL, 0, NULL, NULL },
	{ "input.L2.current.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.5.%i.1.2",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "L2.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.5.%i.1.2",
		NULL, SU_FLAG_OK, &marlin_threshold_current_alarms_info[0], NULL },
	{ "input.L2.current.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.6.%i.1.2",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L2.current.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.7.%i.1.2",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L2.current.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.8.%i.1.2",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L2.current.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.9.%i.1.2",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L3.current", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.4.%i.1.3",
		NULL, 0, NULL, NULL },
	{ "input.L3.current.nominal", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.3.%i.1.3",
		NULL, 0, NULL, NULL },
	{ "input.L3.current.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.5.%i.1.3",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "L3.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.5.%i.1.3",
		NULL, SU_FLAG_OK, &marlin_threshold_current_alarms_info[0], NULL },
	{ "input.L3.current.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.6.%i.1.3",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L3.current.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.7.%i.1.3",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L3.current.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.8.%i.1.3",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	{ "input.L3.current.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.3.3.1.9.%i.1.3",
		NULL, SU_FLAG_NEGINVALID, NULL, NULL },
	/* Sum of all phases realpower, valid for Shark 1ph/3ph only */
	{ "input.realpower", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.5.1.4.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_UNIQUE | SU_FLAG_OK, NULL, NULL },
	/* Fallback 1: Sum of all phases realpower, valid for Marlin 3ph only */
	{ "input.realpower", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.4.%i.1.4",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_UNIQUE | SU_FLAG_OK, NULL, NULL },
	/* Fallback 2: Sum of the phase realpower, valid for Marlin 1ph only */
	{ "input.realpower", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.4.%i.1.2",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "input.L1.realpower", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.4.%i.1.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "input.L2.realpower", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.4.%i.1.2",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "input.L3.realpower", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.4.%i.1.3",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	/* Sum of all phases apparent power, valid for Shark 1ph/3ph only */
	{ "input.power", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.5.1.3.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_UNIQUE | SU_FLAG_OK, NULL, NULL },
	/* Fallback 1: Sum of all phases realpower, valid for Marlin 3ph only */
	{ "input.power", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.3.%i.1.4",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_UNIQUE | SU_FLAG_OK, NULL, NULL },
	/* Fallback 2: Sum of the phase realpower, valid for Marlin 1ph only */
	{ "input.power", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.3.%i.1.2",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "input.L1.power", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.3.%i.1.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "input.L2.power", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.3.%i.1.2",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "input.L3.power", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.3.%i.1.3",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },

	/* Ambient collection */
	{ "ambient.present", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.7.1.1.3.%i.1",
		NULL, SU_FLAG_OK, &marlin_ambient_presence_info[0], NULL },
	{ "ambient.temperature.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.7.1.1.5.%i.1",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "ups.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.7.1.1.5.%i.1",
		NULL, SU_FLAG_OK, &marlin_threshold_temperature_alarms_info[0], NULL },
	{ "ambient.temperature", 0, 0.1, ".1.3.6.1.4.1.534.6.6.7.7.1.1.4.%i.1",
		NULL, SU_FLAG_OK, NULL, NULL },
	/* Low and high threshold use the respective critical levels */
	{ "ambient.temperature.low", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.1.1.7.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.temperature.low.critical", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.1.1.7.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.temperature.low.warning", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.1.1.6.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.temperature.high", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.1.1.9.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.temperature.high.warning", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.1.1.8.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.temperature.high.critical", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.1.1.9.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.humidity.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.7.2.1.5.%i.1",
		NULL, SU_FLAG_OK, &marlin_threshold_status_info[0], NULL },
	{ "ups.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.7.2.1.5.%i.1",
		NULL, SU_FLAG_OK, &marlin_threshold_humidity_alarms_info[0], NULL },
	{ "ambient.humidity", 0, 0.1, ".1.3.6.1.4.1.534.6.6.7.7.2.1.4.%i.1",
		NULL, SU_FLAG_OK, NULL, NULL },
	/* Low and high threshold use the respective critical levels */
	{ "ambient.humidity.low", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.2.1.7.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.humidity.low.warning", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.2.1.6.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.humidity.low.critical", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.2.1.7.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.humidity.high", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.2.1.9.%i.1", NULL,
		SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.humidity.high.warning", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.2.1.8.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	{ "ambient.humidity.high.critical", ST_FLAG_RW, 0.1,
		".1.3.6.1.4.1.534.6.6.7.7.2.1.9.%i.1",
		NULL, SU_FLAG_NEGINVALID | SU_FLAG_OK, NULL, NULL },
	/* Dry contacts on TH module */
	{ "ambient.contacts.1.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.7.3.1.4.%i.1",
		NULL, SU_FLAG_OK, &marlin_ambient_drycontacts_info[0], NULL },
	{ "ambient.contacts.2.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.7.3.1.4.%i.2",
		NULL, SU_FLAG_OK, &marlin_ambient_drycontacts_info[0], NULL },

	/* Outlet collection */
	{ "outlet.id", 0, 1, NULL,
		"0", SU_FLAG_STATIC | SU_FLAG_ABSENT | SU_FLAG_OK, NULL, NULL },
	{ "outlet.desc", ST_FLAG_RW | ST_FLAG_STRING, 20, NULL, "All outlets",
		SU_FLAG_STATIC | SU_FLAG_ABSENT | SU_FLAG_OK, NULL, NULL },
	{ "outlet.count", 0, 1, ".1.3.6.1.4.1.534.6.6.7.1.2.1.22.%i",
		"0", SU_FLAG_STATIC | SU_FLAG_OK, NULL, NULL },
	/* The below ones are the same as the input.* equivalent */
	/* FIXME: transition period, TO BE REMOVED, moved to input.* */
	{ "outlet.frequency", 0, 0.1, ".1.3.6.1.4.1.534.6.6.7.3.1.1.3.%i.1",
		NULL, 0, NULL, NULL },
	{ "outlet.voltage", 0, 0.001, ".1.3.6.1.4.1.534.6.6.7.3.2.1.3.%i.1.1",
		NULL, 0, NULL, NULL },
	{ "outlet.current", 0, 0.01, ".1.3.6.1.4.1.534.6.6.7.3.3.1.4.%i.1.1",
		NULL, 0, NULL, NULL },
	{ "outlet.realpower", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.4.%i.1.4",
		NULL, 0, NULL, NULL },
	{ "outlet.power", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.3.4.1.3.%i.1.4",
		NULL, 0, NULL, NULL },

	/* outlet template definition
	 * Indexes start from 1, ie outlet.1 => <OID>.1 */
	/* Note: the first definition is used to determine the base index (ie 0 or 1) */
	/* outletName: Outlet friendly name, which can be modified by the user */
	{ "outlet.%i.desc", ST_FLAG_RW | ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.1.1.3.%i.%i",
		NULL, SU_FLAG_STATIC | SU_FLAG_OK | SU_OUTLET | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.%i.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.6.1.2.%i.%i",
		NULL, SU_FLAG_OK | SU_OUTLET | SU_TYPE_DAISY_1,
		&marlin_outlet_status_info[0], NULL },
	/* Numeric identifier of the outlet, tied to the whole unit */
	{ "outlet.%i.id", 0, 1, NULL, "%i",
		SU_FLAG_STATIC | SU_FLAG_ABSENT | SU_FLAG_OK | SU_OUTLET | SU_TYPE_DAISY_1,
		NULL, NULL },
	/* outletID: Outlet physical name, related to its number in the group
	 * ex: first outlet of the second group (B) is B1 */
	{ "outlet.%i.name", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.1.1.2.%i.%i",
		NULL, SU_FLAG_STATIC | SU_FLAG_OK | SU_OUTLET | SU_TYPE_DAISY_1,
		NULL, NULL },
	/* FIXME: the last part of the OID gives the group number (i.e. %i.1 means "group 1")
	 * Need to address that, without multiple declaration (%i.%i, SU_OUTLET | SU_OUTLET_GROUP)? */
	{ "outlet.%i.groupid", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.2.1.3.%i.%i.1",
		NULL, SU_FLAG_STATIC | SU_FLAG_UNIQUE | SU_OUTLET | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.%i.groupid", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.2.1.3.%i.%i.2",
		NULL, SU_FLAG_STATIC | SU_FLAG_UNIQUE | SU_OUTLET | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.%i.groupid", ST_FLAG_STRING, SU_INFOSIZE,
		 ".1.3.6.1.4.1.534.6.6.7.6.2.1.3.%i.%i.3",
		NULL, SU_FLAG_STATIC | SU_FLAG_UNIQUE | SU_OUTLET | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.%i.groupid", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.2.1.3.%i.%i.4",
		NULL, SU_FLAG_STATIC | SU_FLAG_UNIQUE | SU_OUTLET | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.%i.groupid", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.2.1.3.%i.%i.5",
		NULL, SU_FLAG_STATIC | SU_FLAG_UNIQUE | SU_OUTLET | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.%i.groupid", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.2.1.3.%i.%i.6",
		NULL, SU_FLAG_STATIC | SU_FLAG_UNIQUE | SU_OUTLET | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.%i.current", 0, 0.001, ".1.3.6.1.4.1.534.6.6.7.6.4.1.3.%i.%i",
		NULL, SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.current.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.4.1.4.%i.%i",
		NULL, SU_OUTLET | SU_TYPE_DAISY_1,
		&marlin_threshold_status_info[0], NULL },
	{ "outlet.%i.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.4.1.4.%i.%i",
		NULL, SU_OUTLET | SU_TYPE_DAISY_1,
		&marlin_threshold_current_alarms_info[0], NULL },
	{ "outlet.%i.current.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.6.4.1.5.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.current.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.6.4.1.6.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.current.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.6.4.1.7.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.current.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.6.4.1.8.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.realpower", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.6.5.1.3.%i.%i",
		NULL, SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.voltage", 0, 0.001, ".1.3.6.1.4.1.534.6.6.7.6.3.1.2.%i.%i",
		NULL, SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.voltage.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.3.1.3.%i.%i",
		NULL, SU_OUTLET | SU_TYPE_DAISY_1,
		&marlin_threshold_status_info[0], NULL },
	{ "outlet.%i.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.3.1.3.%i.%i",
		NULL, SU_OUTLET | SU_TYPE_DAISY_1,
		&marlin_threshold_voltage_alarms_info[0], NULL },
	{ "outlet.%i.voltage.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.6.3.1.4.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.voltage.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.6.3.1.5.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.voltage.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.6.3.1.6.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.voltage.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.6.3.1.7.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.power", 0, 1.0, ".1.3.6.1.4.1.534.6.6.7.6.5.1.2.%i.%i",
		NULL, SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	/* FIXME: handle non switchable units (only measurements), which do not expose this OID */
	{ "outlet.%i.switchable", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.6.1.3.%i.%i",
		"no", SU_FLAG_STATIC | SU_OUTLET | SU_FLAG_OK | SU_TYPE_DAISY_1,
		&marlin_outlet_switchability_info[0], NULL },
	{ "outlet.%i.type", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.6.1.1.5.%i.%i",
		"unknown", SU_FLAG_STATIC | SU_OUTLET | SU_TYPE_DAISY_1,
		&marlin_outlet_type_info[0], NULL },


	/* TODO: handle statistics
	 * outletWh.0.1
	 * outletWhTimer.0.1
	 */

	/* Outlet groups collection */
	{ "outlet.group.count", 0, 1, ".1.3.6.1.4.1.534.6.6.7.1.2.1.21.%i",
		"0", SU_FLAG_STATIC | SU_TYPE_DAISY_1, NULL, NULL },
	/* outlet groups template definition
	 * Indexes start from 1, ie outlet.group.1 => <OID>.1 */
	/* Note: the first definition is used to determine the base index (ie 0 or 1) */
	/* groupID.0.1 = OctetString: A */
	{ "outlet.group.%i.id", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.5.1.1.2.%i.%i",
		NULL, SU_FLAG_STATIC | SU_OUTLET_GROUP | SU_TYPE_DAISY_1, NULL, NULL },
	/* groupName.0.1 = OctetString: Factory Group 1 */
	/* FIXME: SU_FLAG_SEMI_STATIC or SU_FLAG_SETTING => refreshed from time to time or upon call to setvar */
	{ "outlet.group.%i.name", ST_FLAG_RW | ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.5.1.1.3.%i.%i",
		NULL, SU_FLAG_STATIC | SU_OUTLET_GROUP | SU_TYPE_DAISY_1, NULL, NULL },
	/* groupType.0.1 = Integer: outletSection  (4) */
	{ "outlet.group.%i.type", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.5.1.1.4.%i.%i",
		NULL, SU_FLAG_STATIC | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		&marlin_outlet_group_type_info[0], NULL },
	{ "outlet.group.%i.phase", 0, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.5.1.1.2.%i.%i",
		NULL, SU_FLAG_STATIC | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		&marlin_outlet_group_phase_info[0], NULL },
	/* groupControlStatus.0.1 = Integer: on  (1) */
	{ "outlet.group.%i.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.5.6.1.2.%i.%i",
		NULL, SU_FLAG_OK | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		&marlin_outletgroups_status_info[0], NULL },
	/* groupChildCount.0.1 = Integer: 12 */
	{ "outlet.group.%i.count", 0, 1,
		".1.3.6.1.4.1.534.6.6.7.5.1.1.6.%i.%i",
		NULL, SU_OUTLET_GROUP | SU_TYPE_DAISY_1, NULL, NULL },
	/* groupVoltage.0.1 = Integer: 243080 */
	{ "outlet.group.%i.voltage", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.3.1.3.%i.%i",
		NULL, SU_OUTLET_GROUP | SU_TYPE_DAISY_1, NULL, NULL },
	/* groupVoltageThStatus.0.1 = Integer: good (0) */
	{ "outlet.group.%i.voltage.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.5.3.1.4.%i.%i",
		NULL, SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		&marlin_threshold_status_info[0], NULL },
	{ "outlet.group.%i.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.5.3.1.4.%i.%i",
		NULL, SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		&marlin_threshold_voltage_alarms_info[0], NULL },
	{ "outlet.group.%i.voltage.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.3.1.5.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.group.%i.voltage.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.3.1.6.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.group.%i.voltage.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.3.1.7.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	{ "outlet.group.%i.voltage.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.3.1.8.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	/* groupCurrent.0.1 = Integer: 0 */
	{ "outlet.group.%i.current", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.4.1.3.%i.%i",
		NULL, SU_OUTLET_GROUP | SU_TYPE_DAISY_1, NULL, NULL },
	/* groupCurrentCapacity.0.1 = Integer: 16000 */
	{ "outlet.group.%i.current.nominal", 0, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.4.1.2.%i.%i",
		NULL, SU_OUTLET_GROUP | SU_TYPE_DAISY_1, NULL, NULL },
	/* groupCurrentThStatus.0.1 = Integer: good  (0) */
	{ "outlet.group.%i.current.status", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.5.4.1.4.%i.%i",
		NULL, SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		&marlin_threshold_status_info[0], NULL },
	{ "outlet.group.%i.alarm", ST_FLAG_STRING, SU_INFOSIZE,
		".1.3.6.1.4.1.534.6.6.7.5.4.1.4.%i.%i",
		NULL, SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		&marlin_threshold_current_alarms_info[0], NULL },
	/* groupCurrentPercentLoad.0.1 = Integer: 0 */
	{ "outlet.group.%i.load", 0, 1.0,
		".1.3.6.1.4.1.534.6.6.7.5.4.1.10.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	/* groupCurrentThLowerWarning.0.1 = Integer: 0 */
	{ "outlet.group.%i.current.low.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.4.1.5.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	/* groupCurrentThLowerCritical.0.1 = Integer: -1 */
	{ "outlet.group.%i.current.low.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.4.1.6.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	/* groupCurrentThUpperWarning.0.1 = Integer: 12800 */
	{ "outlet.group.%i.current.high.warning", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.4.1.7.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	/* groupCurrentThUpperCritical.0.1 = Integer: 16000 */
	{ "outlet.group.%i.current.high.critical", ST_FLAG_RW, 0.001,
		".1.3.6.1.4.1.534.6.6.7.5.4.1.8.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	/* groupWatts.0.1 = Integer: 2670 */
	{ "outlet.group.%i.realpower", 0, 1.0,
		".1.3.6.1.4.1.534.6.6.7.5.5.1.3.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },
	/* groupVA.0.1 = Integer: 3132 */
	{ "outlet.group.%i.power", 0, 1.0,
		".1.3.6.1.4.1.534.6.6.7.5.5.1.2.%i.%i",
		NULL, SU_FLAG_NEGINVALID | SU_OUTLET_GROUP | SU_TYPE_DAISY_1,
		NULL, NULL },

	/* instant commands. */
	/* Notes:
	 * - load.cycle might be replaced by / mapped on shutdown.reboot
	 * - outletControl{Off,On,Reboot}Cmd values:
	 * 		0-n : Timer
	 * 		-1 : Cancel
	 * 		we currently use "0", so instant On | Off | Reboot... */
	/* no counterpart found!
	{ "outlet.load.off", 0, DO_OFF, AR_OID_OUTLET_STATUS ".0",
		NULL, SU_TYPE_CMD, NULL, NULL },
	{ "outlet.load.on", 0, DO_ON, AR_OID_OUTLET_STATUS ".0",
		NULL, SU_TYPE_CMD, NULL, NULL },
	{ "outlet.load.cycle", 0, DO_CYCLE, AR_OID_OUTLET_STATUS ".0",
		NULL, SU_TYPE_CMD, NULL, NULL }, */

	/* TODO: handle delays */
	{ "outlet.%i.load.off", 0, 0, ".1.3.6.1.4.1.534.6.6.7.6.6.1.3.%i.%i",
		NULL, SU_TYPE_CMD | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.load.on", 0, 0, ".1.3.6.1.4.1.534.6.6.7.6.6.1.4.%i.%i",
		NULL, SU_TYPE_CMD | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },
	{ "outlet.%i.load.cycle", 0, 0, ".1.3.6.1.4.1.534.6.6.7.6.6.1.5.%i.%i",
		NULL, SU_TYPE_CMD | SU_OUTLET | SU_TYPE_DAISY_1, NULL, NULL },

	/* TODO: handle delays
	 * 0-n :Time in seconds until the group command is issued
	 * -1:Cancel a pending group-level Off/On/Reboot command */
	/* groupControlOffCmd.0.1 = Integer: -1 */
	{ "outlet.group.%i.load.off", 0, 0,
		".1.3.6.1.4.1.534.6.6.7.5.6.1.3.%i.%i",
		NULL, SU_TYPE_CMD | SU_OUTLET_GROUP | SU_TYPE_DAISY_1, NULL, NULL },
	/* groupControl0nCmd.0.1 = Integer: -1 */
	{ "outlet.group.%i.load.on", 0, 0,
		".1.3.6.1.4.1.534.6.6.7.5.6.1.4.%i.%i",
		NULL, SU_TYPE_CMD | SU_OUTLET_GROUP | SU_TYPE_DAISY_1, NULL, NULL },
	/* groupControlRebootCmd.0.1 = Integer: -1 */
	{ "outlet.group.%i.load.cycle", 0, 0,
		".1.3.6.1.4.1.534.6.6.7.5.6.1.5.%i.%i",
		NULL, SU_TYPE_CMD | SU_OUTLET_GROUP | SU_TYPE_DAISY_1, NULL, NULL },

// FIXME: miss load.{on,off}.delay

	/* end of structure. */
	{ NULL, 0, 0, NULL, NULL, 0, NULL, NULL }
};


mib2nut_info_t	eaton_marlin = { "eaton_epdu", EATON_MARLIN_MIB_VERSION, NULL, EATON_MARLIN_OID_MODEL_NAME, eaton_marlin_mib, EATON_MARLIN_SYSOID };
